import java.util.Random;

/**
 *  The <code>MazeCell</code> class stores information about each individual cell
 *  in the maze.  The boolean values <code>north</code>, <code>east</code>,
 *  <code>west</code>, and <code>south</code> store which walls are up; e.g., if
 *  <code>north</code> is <code>true</code>, then the north wall is up.
 *  
 *  Each cell also has pointer to its four <code>MazeCell</code> neighbors,
 *  <code>neighborN</code>, <code>neighborE</code>, <code>neighborS</code>, and
 *  <code>neighborW</code>.  If any of these values are null, it means this cell
 *  is on the border of the maze.  
 *
 *  @author Neva Cherniavsky
 *
 */
public class MazeCell {
    
    private boolean north, east, south, west;
    private boolean visited, examined;
    private MazeCell neighborN, neighborE, neighborS, neighborW;
    private Random generator;

    private MazeCell parent;

    /** 
     *  Sets all the walls to <code>true</code> and initializes
     *  the random number generator.
     */
    public MazeCell() {
        north = true;
        east  = true;
        south = true;
        west  = true;
        generator = new Random();
        visited = false;
        examined = false;
        parent = null;
    }

    /**
     *  Sets the visited flag to <code>true</code>.
     */
    public void visit() {
        visited = true;
    }

    /**
     *  Returns whether or not this cell has been visited.
     *  @return <code>true</code> if the cell has been visited.
     */
    public boolean visited() {
        return visited;
    }

    /**
     *  Sets the examined flag to <code>true</code>.
     */
    public void examine() {
        examined = true;
    }

    /**
     *  Returns whether or not this cell has been examined.
     *  @return <code>true</code> if the cell has been examined.
     */
    public boolean examined() {
        return examined;
    }
    
    /**
     *  Sets the pointers to the neighbors of this cell.  If a pointer 
     *  is null, that means this cell is along the border of the maze.
     *  @param n  The north neighbor of this cell.
     *  @param e  The east neighbor of this cell.
     *  @param s  The south neighbor of this cell.
     *  @param w  The west neighbor of this cell.
     */
    public void setNeighbors(MazeCell n, MazeCell e, MazeCell s, MazeCell w) {
        neighborN = n;
        neighborE = e;
        neighborS = s;
        neighborW = w;
    }

    /**
     *  Sets whether or not there are walls up to the north, east, south, and 
     *  west of this cell.
     *  @param north <code>true</code> if there's a wall on the north side of the cell.
     *  @param east <code>true</code> if there's a wall on the east side of the cell.
     *  @param south <code>true</code> if there's a wall on the south side of the cell.     
     *  @param west <code>true</code> if there's a wall on the west side of the cell.
     */
    public void setWalls(boolean north, boolean east, boolean south, boolean west) {
        this.north = north;
        this.east = east;
        this.south = south;
        this.west = west;
    }
    
    /**
     *  Returns whether or not this cell has all its walls up.
     *  @return <code>true</code> if all walls are up.
     */
    public boolean hasAllWalls() {
        return (north && east && south && west);
    }

    /**
     *  Returns whether or not this cell has its north wall up.
     *  @return <code>true</code> if the cell's north wall is up.
     */
    public boolean north() {
        return north;
    }

    /**
     *  Returns whether or not this cell has its south wall up.
     *  @return <code>true</code> if the cell's south wall is up.
     */
    public boolean south() {
        return south;
    }

    /**
     *  Returns whether or not this cell has its east wall up.
     *  @return <code>true</code> if the cell's east wall is up.
     */
    public boolean east() {
        return east;
    }

    /**
     *  Returns whether or not this cell has its west wall up.
     *  @return <code>true</code> if the cell's west wall is up.
     */
    public boolean west() {
        return west;
    }

    /**
     *  Gets the neighbors of this cell.
     *  Returns an array of those neighbors.  The length of the array
     *  is the number of neighbors this cell has.
     *  @return  An array with the neighbors contained within it.
     */
    public MazeCell[] getNeighbors() {
        int count = 0;
        if (!north) count++;
        if (!east) count++;
        if (!south) count++;
        if (!west) count++;
        
        MazeCell n[] = new MazeCell[count];
        count = 0;
        if (!north) {
            n[count] = neighborN;
            count++;
        }
        if (!east) {
            n[count] = neighborE;
            count++;
        }
        if (!south) {
            n[count] = neighborS;
            count++;
        }
        if (!west) {
            n[count] = neighborW;
            count++;   
        }
        return n;
    }

    /**
     *  Knocks down the wall between this cell and the neighbor cell.
     *  The neighbor cell must actually be a neighbor of this cell.
     *  This method is used in conjunction with <code>neighborWithWalls</code>.
     *  @param neighbor  The neighboring cell; must be one of the neighbors
     *                   set in <code>setNeighbors</code>.
     */
    public void knockDownWall(MazeCell neighbor) {
        if (neighbor == neighborN) {
            north = false;
            neighborN.south = false;
        }
        else if (neighbor == neighborS) {
            south = false;
            neighborS.north = false;
        }
        else if (neighbor == neighborE) {
            east = false;
            neighborE.west = false;
        }
        else if (neighbor == neighborW) {
            west = false;
            neighborW.east = false;
        }
        else System.err.println("Called knockDownWall with non-neighbor.");
    }

    /**
     *  Returns a neighbor that has all its walls intact.
     *  @return Neighbor with all its walls intact.
     */
    public MazeCell neighborWithWalls() {
        // Count the number of neighbors with walls.
        int count = 0;
        
        if (neighborN != null && neighborN.hasAllWalls()) count++;
        if (neighborE != null && neighborE.hasAllWalls()) count++;
        if (neighborS != null && neighborS.hasAllWalls()) count++;
        if (neighborW != null && neighborW.hasAllWalls()) count++;

        // If none, return null.
        if (count == 0) return null;
        else {
            // Create array of MazeCells to hold the neighbors with walls.
            MazeCell neighbors[] = new MazeCell[count];
            count = 0;
            if (neighborN != null && neighborN.hasAllWalls()) {
                neighbors[count] = neighborN;
                count++;
            }
            if (neighborE != null && neighborE.hasAllWalls()) {
                neighbors[count] = neighborE;
                count++;
            }
            if (neighborS != null && neighborS.hasAllWalls()) {
                neighbors[count] = neighborS;
                count++;
            }
            if (neighborW != null && neighborW.hasAllWalls()) {
                neighbors[count] = neighborW;
                count++;
            }
            // Randomly choose a neighbor from among 
            // the neighbors with walls, and return it.
            int rand = generator.nextInt(count);
            return neighbors[rand];
        }
    }
}
