import java.util.*;
import java.io.*;

/**
 * MazeFactory is a Factory Design Pattern that
 * generates a new Maze from a text file.  
 * It currently just exits if it hits a parse error.
 * It also contains a method for writing the maze to a file. 
 *
 * @author Albert J. Wong
 * @author Neva Cherniavsky
 */
public class MazeFactory {
    /**
     * parseMaze takes a file in and tries to generate a
     * Maze from that file.
     *
     * @param filename  Name of file with data to
     *  generate a SquareCellMaze.
     * @return Returns a new SquareCellMaze based on the
     *  given filename.
     *
     * @exception FileNotFoundException Thrown if the file
     *  cannot be found.
     * @exception IOException Thrown on a generic IO
     *  Exception
     */
    static Maze parseMaze(String filename) throws FileNotFoundException, 
                                                  IOException {
        BufferedReader br = new BufferedReader(new FileReader(filename) );

        String topLine;
        String midLine;
        String bottomLine;
        Maze maze = null;
        int rows = 0;
        int cols = 0;

        try {
            String line = br.readLine();
            StringTokenizer st = new StringTokenizer(line);
            cols = Integer.parseInt(st.nextToken());
            rows = Integer.parseInt(st.nextToken());
            maze = new Maze(rows, cols);
            
        } catch (NoSuchElementException nsee) {
            System.err.println("Error in Maze: line 1");
            System.err.println(nsee);
            System.exit(1);
        } catch (NumberFormatException nfe) {
            System.err.println("Error in Maze: line 1");
            System.err.println(nfe);
            System.exit(1);
        }
        
        topLine = br.readLine();
        for(int x=0; x < rows; x++) {
            midLine = br.readLine();
            bottomLine = br.readLine();
            
            if( topLine.length() != cols*2+1
                || midLine.length() != cols*2+1
                || bottomLine.length() != cols*2+1 )
                {
                    System.err.println("Error in Maze"
                                       + " near line " + x*2);
                    System.exit(1);
                }
            
            if( topLine.charAt(0) != '+'
                || midLine.charAt(0) != '|'
                || bottomLine.charAt(0) != '+' )
                {
                    System.err.println("Error in Maze (corrupt border) near line "
                                       + x*2 + " col 0");
                    System.exit(1);
                }
            
            if( topLine.charAt(cols*2) != '+'
                || midLine.charAt(cols*2) != '|'
                || bottomLine.charAt(cols*2) != '+' )
                {
                    System.err.println("Error in Maze (corrupt border) "
                                       +"near line " + x*2 + " col "
                                       + topLine.length() );
                    System.exit(1);
                }
            
            for(int y = 0; y < cols; y++) {
                boolean isStart = false;
                boolean isDonut = false;
                boolean northWall = false;
                boolean eastWall = false;
                boolean southWall = false;
                boolean westWall = false;
                MazeCell cell;
                
                int mid = y*2+1;
                int right = mid + 1;
                int left = mid - 1;

                if( x == 0 ) {
                    if( topLine.charAt(left) != '+'
                        || topLine.charAt(mid) != '-'
                        || topLine.charAt(right) != '+' )
                        {
                            System.err.println("Error in Maze "
                                               + "(Corrupt Border) near line "
                                               + x*2 + " col " + mid );
                            System.exit(1);
                        }
                }
                
                if( x == rows-1 ) {
                    if( bottomLine.charAt(left) != '+'
                        || bottomLine.charAt(mid) != '-'
                        || bottomLine.charAt(right) != '+' )
                        {
                            System.err.println("Error in Maze "
                                               +"(corrupt border) near line "
                                               + x*2 + " col " + mid );
                            System.exit(1);
                        }
                }
                
                if(topLine.charAt(mid) == '-') {
                    northWall = true;
                }
                
                if(bottomLine.charAt(mid) == '-') {
                    southWall = true;
                }
                
                if(midLine.charAt(left) == '|') {
                    westWall = true;
                }
                
                if(midLine.charAt(right) == '|') {
                    eastWall = true;
                }
                
                switch(midLine.charAt(mid)) {
                case '*':
                    isStart = true;
                    break;
                    
                case 'O':
                    isDonut = true;
                    break;
                    
                case ' ':
                    // Do nothing special.
                    break;
                    
                default:
                    System.err.println("Error in Maze near line "
                                       + x*2 + " col " + mid );
                    System.exit(1);
                }

                // Create the cell here.
                maze.getCell(x,y).setWalls(northWall, eastWall, southWall, westWall);

                if(isStart) {
                    if(maze.getStartCell() != null) {
                        System.err.println( "Multiple start cells "
                                            + "near line " + x*2 + " col "
                                            + topLine.length() );
                        System.exit(1);
                    }
                    
                    maze.setStartCell(maze.getCell(x,y));
                }
                
                if(isDonut) {
                    if(maze.getEndCell() != null) {
                        System.err.println( "Multiple end cells "
                                            + "near line " + x*2 + " col "
                                            + topLine.length() );
                        System.exit(1);
                    }
                    
                    maze.setEndCell(maze.getCell(x,y));
                }
            }
            
            topLine = bottomLine;
        }
        
        if(maze.getStartCell() == null ) {
            System.err.println("No start cell");
            System.exit(1);
        }
        
        if(maze.getEndCell() == null ) {
            System.err.println("No end cell");
            System.exit(1);
        }
        
        return maze;
    }

    /**
     *  Prints the maze to the file.
     *  @param maze     Maze to print to the file..
     *  @param filename File to print to.
     */
    public static void printMaze(Maze maze, String filename) throws FileNotFoundException,
                                                                    IOException {
        PrintStream out = new PrintStream(new FileOutputStream(filename));

        int cols = maze.getCols();
        int rows = maze.getRows();
        MazeCell cell;
        out.println(cols + " " + rows);
        out.print("+");
        for (int j=0; j<cols; j++) 
            out.print("-+");
        out.println();
        for (int i=0; i<rows; i++) {
            for (int j=0; j<cols; j++) {
                cell = maze.getCell(i,j);
                if (cell.west()) out.print("|");
                else out.print(" ");
                if (cell == maze.getStartCell()) 
                    out.print("*");
                else if (cell == maze.getEndCell())
                    out.print("O");
                else
                    out.print(" ");
                if (j == cols-1) 
                    out.print("|");
            }
            out.println();
            for (int j=0; j<cols; j++) {
                cell = maze.getCell(i,j);
                if (cell.south()) out.print("+-");
                else out.print("+ ");
            }
            out.println("+");
        }
    }
}
