import java.awt.*;
import java.awt.geom.*;
import javax.swing.*;
import java.awt.event.*;
import java.util.Random;

/**
 *  Visualization panel for the maze.  Contains the 
 *  driver for calling various maze solving and maze
 *  formation algorithms.
 */
public class MazeViewer extends JPanel {

    private Maze maze;
    private MazeCell currentCell;

    private final static int pad = 50;
    private final static int size = 25;
    private final static BasicStroke stroke = new BasicStroke(2.0f);

    /**
     *  Main driver function for the program.  Calls various
     *  maze solving and maze formation algorithms, depending 
     *  on the input flags.
     *
     *  @param args  The arguments sent to the method via the
     *               console.  
     */
    public static void main(String args[]) {
        JFrame  frame     = new JFrame("Maze");
        MazeViewer viewer = new MazeViewer();
        frame.getContentPane().add(viewer);
        viewer.init(args);
        frame.setSize(viewer.getSize());
        frame.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
        frame.setVisible(true);
        if (args.length > 2)
            viewer.viewMaze(args[0], args[1], args[2]);
        else 
            viewer.viewMaze(args[0], args[1], null);
    }

    /**
     *  Initializes the MazeViewer by initializing the
     *  maze according to the input arguments.  Checks
     *  that the arguments correspond to expected values.
     *  Parses the maze from the input filename.
     *
     *  @param args  Arguments sent from the console.
     */
    public void init(String[] args) {
        if (!checkArgs(args)) {
            usage();
            System.exit(1);
        }

        if (!(args[0].equals("kruskal") || args[0].equals("dfs"))) {
            try {
                maze = MazeFactory.parseMaze(args[0]);
            }
            catch (Exception e) {
                System.err.println("File " + args[0] + " corrupted or not found.");
                System.exit(1);
            }
        }
        else {
            // Make a 10x10 maze.  Changing these numbers changes the
            // size of the maze.
            maze = new Maze(10, 10);
        }

        maze.setViewer(this);

        setSize(maze.getCols()*size+2*pad, maze.getRows()*size+2*pad);

        setBackground(Color.white);
        setForeground(Color.black);
    }


    /**
     *  Calls the appropriate methods to generate the maze and 
     *  solve the maze. The third parameter may be null.
     *  @param generate  The generation method.  If not equal to "dfs"
     *                   or "kruskal" (i.e. is a filename), not used.
     *  @param solve     The solving method.  Guaranteed to be one of
     *                   "dfs", "bfs", "best2", "best3", or "random".
     *  @param write     The filename to save the maze to.  Can be null,
     *                   if the user did not choose to save the maze.
     */
    public void viewMaze(String generate, String solve, String write) {
        if (generate.equals("kruskal") || generate.equals("dfs"))
            maze.generateMaze(generate);
        maze.solveMaze(solve);
        if (write != null) {
            try {
                MazeFactory.printMaze(maze, write);
            }
            catch (Exception e) {
                System.err.println("Unable to write maze to " + write + ".");
                System.exit(1);
            }
        }
    }

    /**
     *  Repaints the panel with the maze.  Sets the current cell
     *  to the parameter; the current cell will be colored.  Then
     *  waits for 2.5 seconds, so the user can observe the visualization.
     *  @param cell  The current cell, to be colored.
     */
    public synchronized void visualize(MazeCell cell) {
        currentCell = cell;
        repaint();      
        try {
            wait(250);
        }
        catch (Exception e) {}
    }

    /**
     *  Paints the maze.  This method is called whenever <code>repaint()</code> is
     *  called.  For each cell in the maze, the walls of the cell that are up are 
     *  drawn.  Then if the cell is the "start" cell for exploring the maze, a green
     *  square is drawn within the cell; if it is the "end" cell, a red square is
     *  drawn within the cell; and if it is the "current" cell the algorithm is 
     *  looking at, a blue square is drawn within the cell.  Finally, all cells that
     *  have been "visited" are colored yellow.
     *
     *  @param g  Graphics object on which to draw the maze.
     */
    public void paint(Graphics g) {
        super.paint(g);
        Graphics2D g2D = (Graphics2D)g;
        // Make the stroke a thick black line
        g2D.setStroke(stroke);
        for (int j=0; j<maze.getCols(); j++) {
            for (int i=0; i<maze.getRows(); i++) {
                // Draw lines from the x-coordinate to the y-coordinate.
                // On the jth column, the x-coordinate is 
                // j times the size of the cell plus the padding.
                // On the ith row, the y-coordinate is 
                // i times the size of the cell plus the padding.
                int x = j*size+pad;
                int y = i*size+pad;
                MazeCell cell = maze.getCell(i, j);

                // Paint the visited cells yellow
                if (cell.visited()) {
                    g2D.setColor(Color.yellow);
                    g2D.fill(new Rectangle2D.Double(x+5, y+5, size-10, size-10));
                    g2D.setColor(Color.black);
                }
                // Paint the examined cells gray
                if (cell.examined()) {
                    g2D.setColor(Color.gray);
                    g2D.fill(new Rectangle2D.Double(x+5, y+5, size-10, size-10));
                    g2D.setColor(Color.black);
                }
                // Draw a blue rectangle within the current cell.
                if (cell == currentCell) {
                    g2D.setColor(Color.blue);
                    g2D.fill(new Rectangle2D.Double(x+5, y+5, size-10, size-10));
                    g2D.setColor(Color.black);
                }
                // Paint the start cell green
                if (cell == maze.getStartCell()) {
                    g2D.setColor(Color.green);
                    g2D.fill(new Rectangle2D.Double(x+5, y+5, size/2, size/2));
                    g2D.setColor(Color.black);
                }
                // Paint the end cell red
                if (cell == maze.getEndCell()) {
                    g2D.setColor(Color.red);
                    g2D.fill(new Rectangle2D.Double(x+5, y+5, size/2, size/2));
                    g2D.setColor(Color.black);
                }

                // Draw the walls of this cell that are up.
                if (cell.north())
                    g2D.draw(new Line2D.Double(x, y, x+size, y));
                if (cell.south())
                    g2D.draw(new Line2D.Double(x, y+size, x+size, y+size));
                if (cell.east())
                    g2D.draw(new Line2D.Double(x+size, y, x+size, y+size));
                if (cell.west())
                    g2D.draw(new Line2D.Double(x, y, x, y+size));
            }
        }
    }

    /**
     *  We expect the arguments to look like this:<br>
     *  <code>java MazeViewer <generation method> <solve method> [filename]</code><br>     
     *  In the initial version, <generation method> is just the input filename that
     *  we use to read the maze.
     *  @param args  Should be args[0] = input filename, args[1] = 
     *               <solve method> = one of "dfs", "bfs", "best2", "best3", or "random",
     *               and the optional args[2] = output filename.
     */
    private boolean checkArgs(String[] args) {
        if (!(args.length == 2 || args.length == 3)) return false;
        if (!(args[1].equals("dfs") || 
              args[1].equals("bfs") || 
              args[1].equals("best2") ||
              args[1].equals("best3") ||
              args[1].equals("random"))) return false;
        return true;
    }

    /**
     *  Prints out a usage message to the error console.
     */
    private void usage() {
        System.err.println("Usage: java MazeViewer <generation method> <solve method> [filename]");
        System.err.println("<generation method> is one of three methods:");
        System.err.println("\tfilename\t Generate the maze from the given file");
        System.err.println("\tdfs\t Generate the maze using depth first search");
        System.err.println("\tkruskal\t Generate the maze using Kruskal's algorithm");
        System.err.println("<solve method> is one of five methods:");
        System.err.println("\trandom\t Solve the maze randomly");
        System.err.println("\tdfs\t Solve the maze using depth first search");
        System.err.println("\tbfs\t Solve the maze using breadth first search");
        System.err.println("\tbest2\t Solve the maze using best first search and a binary heap");
        System.err.println("\tbest3\t Solve the maze using best first search and a 3-heap");
        System.err.println("You may also save the generated maze to a file using the filename option.");
    }
}



